/*
    Copyright (c) 2011, BogDan Vatra <bog_dan_ro@yahoo.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFile>
#include <QFont>
#include <QApplication>
#include <QTime>
#include <QDebug>
#include "androidstyle.h"
#include "json.h"
#include <qdrawutil.h>
#include <QPixmapCache>
#include <QFileInfo>
#include <QStyleOption>
#include <QPainter>


namespace {
    const int textStyle_bold = 1;
    const int textStyle_italic = 2;

    const int typeface_sans = 1;
    const int typeface_serif = 2;
    const int typeface_monospace = 3;

    const quint32 NO_COLOR = 1;
    const quint32 TRANSPARENT_COLOR = 0;

    int calculateStretch(int boundsLimit, int startingPoint,
                              int srcSpace, int numStrechyPixelsRemaining,
                              int numFixedPixelsRemaining) {
        int spaceRemaining = boundsLimit - startingPoint;
        int stretchySpaceRemaining = spaceRemaining - numFixedPixelsRemaining;
        return ((float)srcSpace * stretchySpaceRemaining / numStrechyPixelsRemaining +.5);
    }

    void drawAndroid9PatchImage(QPainter * painter, const QRect & bounds, const AndroidStyle::Android9PatchChunk & chunk, const QPixmap & pixmap)
    {
        // shamelessly stolen from Android's sources (NinepatchImpl.cpp) and adapted for Qt
        const int pixmapWidth = pixmap.width();
        const int pixmapHeight = pixmap.height();

        if (bounds.isNull() || !pixmapWidth || !pixmapHeight)
            return;

        QPainter::RenderHints savedHints=painter->renderHints();

        // The patchs doesn't need smooth transform !
        painter->setRenderHints(QPainter::SmoothPixmapTransform, false);


        QRectF dst;
        QRectF src;

        const int32_t x0 = chunk.xDivs[0];
        const int32_t y0 = chunk.yDivs[0];
        const quint8 numXDivs = chunk.xDivs.size();
        const quint8 numYDivs = chunk.yDivs.size();
        int i;
        int j;
        int colorIndex = 0;
        quint32 color;
        bool xIsStretchable;
        const bool initialXIsStretchable =  (x0 == 0);
        bool yIsStretchable = (y0 == 0);
        const int bitmapWidth = pixmap.width();
        const int bitmapHeight = pixmap.height();

        int * dstRights = (int*) alloca((numXDivs + 1) * sizeof(int));
        bool dstRightsHaveBeenCached = false;

        int numStretchyXPixelsRemaining = 0;
        for (i = 0; i < numXDivs; i += 2) {
            numStretchyXPixelsRemaining += chunk.xDivs[i + 1] - chunk.xDivs[i];
        }
        int numFixedXPixelsRemaining = bitmapWidth - numStretchyXPixelsRemaining;
        int numStretchyYPixelsRemaining = 0;
        for (i = 0; i < numYDivs; i += 2) {
            numStretchyYPixelsRemaining += chunk.yDivs[i + 1] - chunk.yDivs[i];
        }
        int numFixedYPixelsRemaining = bitmapHeight - numStretchyYPixelsRemaining;
        src.setTop(0);
        dst.setTop(bounds.top());
        // The first row always starts with the top being at y=0 and the bottom
        // being either yDivs[1] (if yDivs[0]=0) of yDivs[0].  In the former case
        // the first row is stretchable along the Y axis, otherwise it is fixed.
        // The last row always ends with the bottom being bitmap.height and the top
        // being either yDivs[numYDivs-2] (if yDivs[numYDivs-1]=bitmap.height) or
        // yDivs[numYDivs-1]. In the former case the last row is stretchable along
        // the Y axis, otherwise it is fixed.
        //
        // The first and last columns are similarly treated with respect to the X
        // axis.
        //
        // The above is to help explain some of the special casing that goes on the
        // code below.

        // The initial yDiv and whether the first row is considered stretchable or
        // not depends on whether yDiv[0] was zero or not.
        for (j = yIsStretchable ? 1 : 0;
             j <= numYDivs && src.top() < bitmapHeight;
              j++, yIsStretchable = !yIsStretchable) {
            src.setLeft(0);
            dst.setLeft(bounds.left());
            if (j == numYDivs) {
                src.setBottom(bitmapHeight);
                dst.setBottom(bounds.bottom());
            } else {
                src.setBottom(chunk.yDivs[j]);
                const int srcYSize = src.bottom() - src.top();
                if (yIsStretchable) {
                    dst.setBottom(dst.top() + calculateStretch(bounds.bottom(), dst.top(),
                                                              srcYSize,
                                                              numStretchyYPixelsRemaining,
                                                              numFixedYPixelsRemaining));
                    numStretchyYPixelsRemaining -= srcYSize;
                } else {
                    dst.setBottom(dst.top() + srcYSize);
                    numFixedYPixelsRemaining -= srcYSize;
                }
            }

            xIsStretchable = initialXIsStretchable;
            // The initial xDiv and whether the first column is considered
            // stretchable or not depends on whether xDiv[0] was zero or not.
            for (i = xIsStretchable ? 1 : 0;
                  i <= numXDivs && src.left() < bitmapWidth;
                  i++, xIsStretchable = !xIsStretchable) {
                color = chunk.colors[colorIndex++];
                if (i == numXDivs) {
                    src.setRight(bitmapWidth);
                    dst.setRight(bounds.right());
                } else {
                    src.setRight(chunk.xDivs[i]);
                    if (dstRightsHaveBeenCached) {
                        dst.setRight(dstRights[i]);
                    } else {
                        const int srcXSize = src.right() - src.left();
                        if (xIsStretchable) {
                            dst.setRight(dst.left() + calculateStretch(bounds.right(), dst.left(),
                                                                      srcXSize,
                                                                      numStretchyXPixelsRemaining,
                                                                      numFixedXPixelsRemaining));
                            numStretchyXPixelsRemaining -= srcXSize;
                        } else {
                            dst.setRight(dst.left() + srcXSize);
                            numFixedXPixelsRemaining -= srcXSize;
                        }
                        dstRights[i] = dst.right();
                    }
                }
                // If this horizontal patch is too small to be displayed, leave
                // the destination left edge where it is and go on to the next patch
                // in the source.
                if (src.left() >= src.right()) {
                    src.setLeft(src.right());
                    continue;
                }
                // Make sure that we actually have room to draw any bits
                if (dst.right() <= dst.left() || dst.bottom() <= dst.top()) {
                    goto nextDiv;
                }
                // If this patch is transparent, skip and don't draw.
                if (color == TRANSPARENT_COLOR)
                    goto nextDiv;
                if (color != NO_COLOR)
                    painter->fillRect(dst, (QRgb)color);
                else
                    painter->drawPixmap(dst, pixmap, src);
    nextDiv:
                src.setLeft(src.right());
                dst.setLeft(dst.right());
            }
            src.setTop(src.bottom());
            dst.setTop(dst.bottom());
            dstRightsHaveBeenCached = true;
        }
        painter->setRenderHints(savedHints);
    }
}


AndroidStyle::AndroidStyle() :
    QCommonStyle()
{
    QTime t;
    t.start();
    QString stylePath(qgetenv("MINISTRO_ANDROID_STYLE_PATH"));
    if (!stylePath.length())
        stylePath=":/android/style/ics/";
    QFile f(stylePath+"style.json");
    if (!f.open(QIODevice::ReadOnly))
        return;
    JsonReader reader;
    if (!reader.parse(f.readAll()))
    {
        qCritical()<<reader.errorString();
        return;
    }
    qDebug()<<"Parsing json file took"<<t.elapsed()<<"ms";
    QVariantMap jsDocument = reader.result().toMap();
    foreach(QString key, jsDocument.keys())
    {
        int itemType=qtControl(key);
        if (-1==itemType)
            continue;
        QVariantMap item=jsDocument.value(key).toMap();
        if(item.contains("qtClass"))
        { // the item has palette and font informations for a specifi Qt Class (e.g. QWidget, QPushButton, etc.)
            const QString qtClassName=item.value("qtClass").toString();

            // extract font information
            QFont font;

            // font size (in pixels)
            QVariantMap::const_iterator it=item.find("TextAppearance_textSize");
            if (it!=item.end())
                font.setPixelSize(it.value().toInt());

            // font style
            it=item.find("TextAppearance_textStyle");
            if (it!=item.end())
            {
                const int style=it.value().toInt();
                font.setBold(style & textStyle_bold);
                font.setItalic(style & textStyle_italic);
            }

            // font typefac
            it=item.find("TextAppearance_typeface");
            if (it!=item.end())
            {
                QFont::StyleHint styleHint=QFont::AnyStyle;
                switch(it.value().toInt())
                {
                    case typeface_sans:
                        styleHint=QFont::SansSerif;
                        break;
                    case typeface_serif:
                        styleHint=QFont::Serif;
                        break;
                    case typeface_monospace:
                        styleHint=QFont::Monospace;
                        break;
                }
                font.setStyleHint(styleHint, QFont::PreferMatch);
            }
            QApplication::setFont(font, qtClassName.toUtf8().constData());
            // extract font information

            // extract palette information
            QPalette palette;
            it=item.find("TextAppearance_textColor");
            if (it!=item.end())
                setPaletteColor(it.value().toMap(), palette, QPalette::WindowText);

            it=item.find("TextAppearance_textColorLink");
            if (it!=item.end())
                setPaletteColor(it.value().toMap(), palette, QPalette::Link);

            it=item.find("TextAppearance_textColorHighlight");
            if (it!=item.end())
                palette.setColor(QPalette::HighlightedText, (QRgb)it.value().toInt());
            palette.setColor(QPalette::Window, Qt::black);
            QApplication::setPalette(palette, qtClassName.toUtf8().constData());
            // extract palette information
        }

        AndroidItemStateInfoList & list = m_androidControlsHash[itemType];
        QVariantMap::const_iterator it=item.find("CompoundButton_button");
        if (it!=item.end())
        {
            setBackground(list, it.value().toMap());
            it=item.find("View_background"); // we need padding info
            if (it!=item.end())
            {
                QVariantMap backgroundInfo(it.value().toMap());
                QString type = backgroundInfo.value("type").toString();
                if (type == "9patch")
                {
                    QMargins margins = extractMargins(backgroundInfo.value("padding").toMap());
                    foreach(AndroidItemStateInfo * back, list)
                        back->padding = margins;
                }
            }
        }
        else
        {
            it=item.find("View_background");
            if (it!=item.end())
                setBackground(list, it.value().toMap());
        }

        it=item.find("View_minWidth");
        if (it!=item.end())
            foreach(AndroidItemStateInfo * back, list)
                back->size.setWidth(it.value().toInt());

        it=item.find("View_minHeight");
        if (it!=item.end())
            foreach(AndroidItemStateInfo * back, list)
                back->size.setHeight(it.value().toInt());

    }
    qDebug()<<"Parsing json file took"<<t.elapsed()<<"ms";
}

AndroidStyle::~AndroidStyle()
{
    foreach(const AndroidItemStateInfoList & backgroundList, m_androidControlsHash)
        foreach(AndroidItemStateInfo * background, backgroundList)
            delete background;
}


bool AndroidStyle::setBackground(AndroidItemStateInfoList & list, const QVariantMap & drawable)
{
    const QString type=drawable.value("type").toString();
    if (type == "9patch")
    {
        AndroidItemStateInfo * style = list.last();
        style->padding = extractMargins(drawable.value("padding").toMap());
        QVariantMap chunkData=drawable.value("chunkInfo").toMap();
        extractIntArray(chunkData.value("xdivs").toList(), style->chunkData.xDivs);
        extractIntArray(chunkData.value("ydivs").toList(), style->chunkData.yDivs);
        extractIntArray(chunkData.value("colors").toList(), style->chunkData.colors);
        return setBackground(list, drawable.value("drawable").toMap());
    }
    if (type == "image")
    {
        AndroidItemStateInfo * style = list.last();
        style->filePath=drawable.value("path").toByteArray();
        style->size.setHeight(drawable.value("height").toInt());
        style->size.setWidth(drawable.value("width").toInt());
        return true;
    }

    if (type == "stateslist")
    {
        QVariantList states=drawable.value("stateslist").toList();
        foreach(QVariant stateVariant, states)
        {
            QVariantMap state=stateVariant.toMap();
            const int s=extractState(state.value("states").toMap());
            if (-1 == s)
                continue;
            AndroidItemStateInfo * style = new AndroidItemStateInfo;
            style->state=s;
            list.push_back(style);
            if (!setBackground(list, state.value("drawable").toMap()))
                return false;
        }
        return true;
    }

    if (type == "layer")
    {
        QVariantList layers=drawable.value("layers").toList();
        foreach(QVariant layer, layers)
            if (!setBackground(list, layer.toMap()))
                return false;
        return true;
    }
    return false;
}

AndroidStyle::AndroidItemStateInfo *AndroidStyle::bestAndroidItemMatch(ItemType itemType, const AndroidItemStateInfoList &list, const QStyleOption *opt)
{
    AndroidItemStateInfo* bestMatch = 0;
    uint bestCost=0xffff;
    foreach(AndroidItemStateInfo* item, list)
    {
        if ((int)opt->state == item->state)
            return item;
        uint cost = 0;

        int difference=(int)opt->state^item->state;
        if ( difference & QStyle::State_Active )
            cost += 1000;

        if ( difference & QStyle::State_Enabled)
            cost += 1000;

        if ( itemType == QC_Button && difference & QStyle::State_Raised )
            cost += 1000;

        if ( itemType == QC_Button && difference & QStyle::State_Sunken )
            cost += 1000;

        if ( difference & QStyle::State_Off )
            cost += 1000;

        if ( difference & QStyle::State_On )
            cost += 1000;

        if ( difference & QStyle::State_HasFocus )
            cost += 1000;

        if ( difference & QStyle::State_Selected )
            cost += 1000;

        if (cost<bestCost)
        {
            bestCost = cost;
            bestMatch = item;
        }
    }
    return bestMatch;
}

void AndroidStyle::drawBackgroud(AndroidItemStateInfo *background, const QStyleOption *opt, QPainter *p,
                          const QWidget */*w*/)
{
    QTime t;
    t.start();
    if (background->hashKey.isEmpty())
        background->hashKey=QFileInfo(background->filePath).fileName().toUtf8();

    QPixmap pm;
    if (!QPixmapCache::find(background->hashKey, &pm))
    {
        pm.load(background->filePath);
        QPixmapCache::insert(background->hashKey, pm);
    }

    if (background->chunkData.colors.size() && background->chunkData.xDivs.size() && background->chunkData.yDivs.size())
        drawAndroid9PatchImage(p, opt->rect, background->chunkData, pm);
    else
    {
        p->drawPixmap(0,(opt->rect.height()-pm.height())/2, pm);
    }
    qDebug()<<"drawBackgroud took "<<t.elapsed()<<"ms"<<" for file "<<background->filePath;
}

int AndroidStyle::extractState(const QVariantMap &value)
{
    int state = QStyle::State_None;
    foreach (const QString key, value.keys())
    {
        bool val=value.value(key).toString()=="true"?true:false;
        if (key=="enabled" && val)
        {
            state |= QStyle::State_Enabled;
            continue;
        }

        if (key=="window_focused" && val)
        {
            state |= QStyle::State_Active;
            continue;
        }

        if (key=="focused" && val)
        {
            state |= QStyle::State_HasFocus;
            continue;
        }

        if (key=="checked")
        {
            state |= (val?QStyle::State_On:QStyle::State_Off);
            continue;
        }

        if (key=="pressed")
        {
            state |= (val?QStyle::State_Raised:QStyle::State_Sunken);
            state |= QStyle::State_Enabled | QStyle::State_HasFocus;
            continue;
        }

        if (key=="selected" && val)
        {
            state |= QStyle::State_Selected;
            state |= QStyle::State_Enabled | QStyle::State_HasFocus;
            continue;
        }

        if (key=="active" && val)
        {
            state |= QStyle::State_Active;
            continue;
        }

        if (key=="backdroud" && val)
            return -1;
    }
    return state;
}

QMargins AndroidStyle::extractMargins(const QVariantMap &value)
{
    QMargins m;
    m.setLeft(value.value("left").toInt());
    m.setRight(value.value("right").toInt());
    m.setTop(value.value("top").toInt());
    m.setBottom(value.value("bottom").toInt());
    return m;
}

void AndroidStyle::extractIntArray(const QVariantList &values, QVector<int> & array)
{
    foreach(QVariant value, values)
        array<<value.toInt();
}

void AndroidStyle::setPaletteColor(const QVariantMap &object, QPalette & palette
                                   , QPalette::ColorRole role)
{
    // QPalette::Active -> ENABLED_FOCUSED_WINDOW_FOCUSED_STATE_SET
    palette.setColor(QPalette::Active, role
                     , (QRgb)object.value("ENABLED_FOCUSED_WINDOW_FOCUSED_STATE_SET").toInt());

    // QPalette::Inactive -> ENABLED_STATE_SET
    palette.setColor(QPalette::Inactive, role
                     , (QRgb)object.value("ENABLED_STATE_SET").toInt());

    // QPalette::Disabled -> EMPTY_STATE_SET
    palette.setColor(QPalette::Disabled, role
                     , (QRgb)object.value("EMPTY_STATE_SET").toInt());

    if (role == QPalette::WindowText)
    {
        // QPalette::BrightText -> PRESSED
        // QPalette::Active -> PRESSED_ENABLED_FOCUSED_WINDOW_FOCUSED_STATE_SET
        palette.setColor(QPalette::Active, QPalette::BrightText
                         , (QRgb)object.value("PRESSED_ENABLED_FOCUSED_WINDOW_FOCUSED_STATE_SET").toInt());

        // QPalette::Inactive -> PRESSED_ENABLED_STATE_SET
        palette.setColor(QPalette::Inactive, QPalette::BrightText
                         , (QRgb)object.value("PRESSED_ENABLED_STATE_SET").toInt());

        // QPalette::Disabled -> PRESSED_STATE_SET
        palette.setColor(QPalette::Disabled, QPalette::BrightText
                         , (QRgb)object.value("PRESSED_STATE_SET").toInt());


        // QPalette::HighlightedText -> SELECTED
        // QPalette::Active -> ENABLED_SELECTED_WINDOW_FOCUSED_STATE_SET
        palette.setColor(QPalette::Active, QPalette::HighlightedText
                         , (QRgb)object.value("ENABLED_SELECTED_WINDOW_FOCUSED_STATE_SET").toInt());

        // QPalette::Inactive -> ENABLED_SELECTED_STATE_SET
        palette.setColor(QPalette::Inactive, QPalette::HighlightedText
                         , (QRgb)object.value("ENABLED_SELECTED_STATE_SET").toInt());

        // QPalette::Disabled -> SELECTED_STATE_SET
        palette.setColor(QPalette::Disabled, QPalette::HighlightedText
                         , (QRgb)object.value("SELECTED_STATE_SET").toInt());


        // Same colors for Text
        palette.setColor(QPalette::Active, QPalette::Text, palette.color(QPalette::Active, QPalette::WindowText));
        palette.setColor(QPalette::Inactive, QPalette::Text, palette.color(QPalette::Inactive, QPalette::WindowText));
        palette.setColor(QPalette::Disabled, QPalette::Text, palette.color(QPalette::Disabled, QPalette::WindowText));

        // And for ButtonText
        palette.setColor(QPalette::Active, QPalette::ButtonText, palette.color(QPalette::Active, QPalette::WindowText));
        palette.setColor(QPalette::Inactive, QPalette::ButtonText, palette.color(QPalette::Inactive, QPalette::WindowText));
        palette.setColor(QPalette::Disabled, QPalette::ButtonText, palette.color(QPalette::Disabled, QPalette::WindowText));
    }
}

AndroidStyle::ItemType AndroidStyle::qtControl(const QString & android)
{
    if (android == "buttonStyle")
        return QC_Button;
    if (android == "editTextStyle")
        return QC_EditText;
    if (android == "radioButtonStyle")
        return QC_RadioButton;
    if (android == "checkboxStyle")
        return QC_Checkbox;
    if (android == "textViewStyle")
        return QC_View;
    if (android == "buttonStyleToggle")
        return QC_Switch;
    if (android == "spinnerStyle")
        return QC_Combobox;

    return QC_UnknownType;
}

AndroidStyle::ItemType AndroidStyle::qtControl(QStyle::ComplexControl control)
{
    switch(control)
    {
        case CC_ComboBox:
            return QC_Combobox;
        case CC_Slider:
            return QC_Slider;
        case CC_GroupBox:
            return QC_View;
        default:
            return QC_UnknownType;
    }
}

AndroidStyle::ItemType AndroidStyle::qtControl(QStyle::ContentsType contentsType)
{
    switch(contentsType)
    {
        case CT_PushButton:
            return QC_Button;
        case CT_CheckBox:
            return QC_Checkbox;
        case CT_RadioButton:
            return QC_RadioButton;
        case CT_ComboBox:
            return QC_Combobox;
        case CT_ProgressBar:
            return QC_ProgressBar;
        case CT_Slider:
            return QC_Slider;
        case CT_TabWidget:
            return QC_Tab;
        case CT_TabBarTab:
            return QC_TabButton;
        case CT_LineEdit:
            return QC_EditText;
        case CT_GroupBox:
            return QC_GroupBox;
        default:
            return QC_UnknownType;
    }
}

AndroidStyle::ItemType AndroidStyle::qtControl(QStyle::ControlElement controlElement)
{
    switch(controlElement)
    {
        case CE_PushButton:
        case CE_PushButtonBevel:
        case CE_PushButtonLabel:
            return QC_Button;

        case CE_CheckBox:
        case CE_CheckBoxLabel:
            return QC_Checkbox;

        case CE_RadioButton:
        case CE_RadioButtonLabel:
            return QC_RadioButton;

        case CE_TabBarTab:
        case CE_TabBarTabShape:
        case CE_TabBarTabLabel:
            return QC_Tab;

        case CE_ProgressBar:
        case CE_ProgressBarGroove:
        case CE_ProgressBarContents:
        case CE_ProgressBarLabel:
            return QC_ProgressBar;

        case CE_ComboBoxLabel:
            return QC_Combobox;

        default:
            return QC_UnknownType;
    }
}

AndroidStyle::ItemType AndroidStyle::qtControl(QStyle::PrimitiveElement primitiveElement)
{
    switch(primitiveElement)
    {
        case QStyle::PE_PanelLineEdit:
        case QStyle::PE_FrameLineEdit:
            return QC_EditText;

        case QStyle::PE_FrameWindow:
        case QStyle::PE_Widget:
        case QStyle::PE_Frame:
        case QStyle::PE_FrameFocusRect:
            return QC_View;
        default:
            return QC_UnknownType;
    }
}

AndroidStyle::ItemType AndroidStyle::qtControl(QStyle::SubElement subElement)
{
    switch(subElement)
    {
        case QStyle::SE_LineEditContents:
            return QC_EditText;

        case QStyle::SE_PushButtonContents:
        case QStyle::SE_PushButtonFocusRect:
            return QC_Button;

        case SE_RadioButtonContents:
            return QC_RadioButton;

        case SE_CheckBoxContents:
            return QC_Checkbox;

        default:
            return QC_UnknownType;
    }
}


void AndroidStyle::drawPrimitive(PrimitiveElement pe, const QStyleOption *opt, QPainter *p,
                           const QWidget *w) const
{
    const ItemType itemType=qtControl(pe);
    AndroidControlsHash::const_iterator it;
    AndroidItemStateInfo * backgroud;
    if ( itemType != QC_UnknownType
            && (it=m_androidControlsHash.find(itemType))!=m_androidControlsHash.end()
            && (backgroud=bestAndroidItemMatch(itemType, it.value(), opt)))

        drawBackgroud(backgroud, opt, p, w);
    else
        QCommonStyle::drawPrimitive(pe, opt, p, w);
}


void AndroidStyle::drawControl(QStyle::ControlElement element, const QStyleOption *opt, QPainter *p,
                         const QWidget *w) const
{
    const ItemType itemType=qtControl(element);
    AndroidControlsHash::const_iterator it;
    AndroidItemStateInfo * backgroud;
    if ( itemType != QC_UnknownType
            && (it=m_androidControlsHash.find(itemType))!=m_androidControlsHash.end()
            && (backgroud=bestAndroidItemMatch(itemType, it.value(), opt)))
    {
        drawBackgroud(backgroud, opt, p, w);
        switch(itemType)
        {
            case QC_Button:
                if (const QStyleOptionButton *buttonOption =
                        qstyleoption_cast<const QStyleOptionButton *>(opt))
                {
                    QStyleOptionButton copy (*buttonOption);
                    copy.rect.adjust(backgroud->padding.left(), backgroud->padding.top(), -backgroud->padding.right(), -backgroud->padding.bottom());
                    QCommonStyle::drawControl(CE_PushButtonLabel, &copy, p, w);
                }
                break;
        case QC_Checkbox:
        case QC_RadioButton:
                if (const QStyleOptionButton *btn =
                        qstyleoption_cast<const QStyleOptionButton *>(opt))
                {
                    const bool isRadio = (element == CE_RadioButton);
                    QStyleOptionButton subopt(*btn);
                    subopt.rect = subElementRect(isRadio ? SE_RadioButtonContents
                                                         : SE_CheckBoxContents, btn, w);
                    QCommonStyle::drawControl(isRadio ? CE_RadioButtonLabel : CE_CheckBoxLabel, &subopt, p, w);
                }
                break;
        }
    }
    else
        QCommonStyle::drawControl(element, opt, p, w);
}

QRect AndroidStyle::subElementRect(SubElement subElement, const QStyleOption *option,
                             const QWidget *widget) const
{
    const ItemType itemType=qtControl(subElement);
    AndroidControlsHash::const_iterator it;
    AndroidItemStateInfo * backgroud;
    if ( itemType != QC_UnknownType
            && (it=m_androidControlsHash.find(itemType))!=m_androidControlsHash.end()
            && (backgroud=bestAndroidItemMatch(itemType, it.value(), option)))
    {
            QRect r = option->rect.adjusted(backgroud->padding.left(), backgroud->padding.top(), -backgroud->padding.right(), -backgroud->padding.bottom());
            return visualRect(option->direction, option->rect, r);
    }
    return QCommonStyle::subElementRect(subElement, option, widget);
}

void AndroidStyle::drawComplexControl(ComplexControl cc, const QStyleOptionComplex *opt, QPainter *p,
                                const QWidget *widget) const
{
    switch(cc)
    {
        default:
            QCommonStyle::drawComplexControl(cc, opt, p, widget);
            break;
    }
}

QStyle::SubControl AndroidStyle::hitTestComplexControl(ComplexControl cc, const QStyleOptionComplex *opt,
                                         const QPoint &pt, const QWidget *widget) const
{
    switch(cc)
    {
        default:
            return QCommonStyle::hitTestComplexControl(cc, opt, pt, widget);
    }
    return QStyle::SC_None;
}

QRect AndroidStyle::subControlRect(ComplexControl cc, const QStyleOptionComplex *opt,
                             SubControl sc, const QWidget *widget) const
{
    const ItemType itemType=qtControl(cc);
    AndroidControlsHash::const_iterator it;
    AndroidItemStateInfo * backgroud;
    if ( itemType != QC_UnknownType
            && (it=m_androidControlsHash.find(itemType))!=m_androidControlsHash.end()
            && (backgroud=bestAndroidItemMatch(itemType, it.value(), opt)))
    {
            QRect r = opt->rect.adjusted(backgroud->padding.left(), backgroud->padding.top(), -backgroud->padding.right(), -backgroud->padding.bottom());
            return visualRect(opt->direction, opt->rect, r);
    }
    return QCommonStyle::subControlRect(cc, opt, sc, widget);
}

int AndroidStyle::pixelMetric(PixelMetric metric, const QStyleOption *option,
                        const QWidget *widget) const
{
    switch(metric)
    {
    case PM_ButtonMargin:
    case PM_FocusFrameVMargin:
    case PM_FocusFrameHMargin:
    case PM_ComboBoxFrameWidth:
    case PM_SpinBoxFrameWidth:
        return 0;
    default:
        return QCommonStyle::pixelMetric(metric, option, widget);
    }

}

QSize AndroidStyle::sizeFromContents(ContentsType ct, const QStyleOption *opt,
                               const QSize &contentsSize, const QWidget *w) const
{
    const ItemType itemType=qtControl(ct);
    AndroidControlsHash::const_iterator it;
    AndroidItemStateInfo * background;
    if ( itemType != QC_UnknownType
            && (it=m_androidControlsHash.find(itemType))!=m_androidControlsHash.end()
            && (background=bestAndroidItemMatch(itemType, it.value(), opt)))
        return background->size;
    return QCommonStyle::sizeFromContents(ct, opt, contentsSize, w);
}

int AndroidStyle::styleHint(StyleHint stylehint, const QStyleOption *opt,
                      const QWidget *widget, QStyleHintReturn* returnData) const
{
    switch (stylehint)
    {
    case SH_RequestSoftwareInputPanel:
        return RSIP_OnMouseClick;
    default:
        return QCommonStyle::styleHint(stylehint, opt, widget, returnData);
    }
}

QPixmap AndroidStyle::standardPixmap(StandardPixmap standardPixmap, const QStyleOption *opt,
                               const QWidget *widget) const
{
    return QCommonStyle::standardPixmap(standardPixmap, opt, widget);
//    return QPixmap(); // should we support any standard pixmaps ?
}

QPixmap AndroidStyle::generatedIconPixmap(QIcon::Mode iconMode, const QPixmap &pixmap,
                                    const QStyleOption *opt) const
{
    return QCommonStyle::generatedIconPixmap(iconMode, pixmap, opt);
}
